# _*_ coding: utf-8 _*_
#
# hspylib v0.11.31
#
# Package: main.hspylib.core.crud.db
"""Package initialization."""

__all__ = [
  'db_repository', 
  'firebase', 
  'mysql', 
  'sql_factory'
]
__version__ = '0.11.31'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.json_search
      @file: json_search.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import re
from typing import Any

from pyparsing import unicode


class JsonSearch:
    """TODO"""

    RE_JSON_NAME = '[a-zA-Z0-9_\\- ]'
    RE_JSON_ARRAY_INDEX = '[0-9]{1,}'

    def __init__(
        self,
        separator='.',
        json_name_re=RE_JSON_NAME,
        json_array_index_re=RE_JSON_ARRAY_INDEX):
        """Construction"""

        self.separator = separator
        self.jsonNameRe = json_name_re
        self.jsonArrayIndexRe = json_array_index_re
        self.pat_elem = None
        self.pat_sel_elem_val = None
        self.pat_sub_expr = None
        self.pat_sub_expr_val = None

    def __find_next_element__(
        self,
        root_element,
        match_name,
        match_value=None,
        fetch_parent=False) -> Any:
        """Find the next element in the list matching the specified value."""

        selected_element = root_element
        if isinstance(selected_element, list):
            for nextInList in root_element:
                if isinstance(nextInList, dict):
                    selected_element = nextInList.get(match_name)
                    if selected_element and (match_value is None or (match_value and selected_element == match_value)):
                        # To return the parent element instead of the leaf
                        if fetch_parent:
                            selected_element = nextInList
                        break
                    selected_element = None
                elif isinstance(nextInList, unicode):
                    selected_element = root_element
                elif isinstance(nextInList, list):
                    selected_element = self.__find_next_element__(nextInList, match_name, match_value)
        elif isinstance(selected_element, dict):
            el = selected_element.get(match_name)
            if el and (match_value is None or (match_value and el == match_value)):
                # To return the parent element instead of the leaf
                if not fetch_parent:
                    selected_element = el
            else:
                selected_element = None
        else:
            selected_element = None

        return selected_element

    def __find_in_subex__(
        self,
        sub_expressions,
        sub_selected_element,
        pat_subst_expr_val,
        fetch_parent=False) -> Any:
        """Find the element in the sub-expressions."""

        for nextSubExpr in sub_expressions:

            if nextSubExpr:
                sub_parts = re.search(pat_subst_expr_val, nextSubExpr)
                sub_elem_id = sub_parts.group(1)
                sub_elem_val = sub_parts.group(3)
                sub_selected_element = self.__find_next_element__(
                    sub_selected_element, sub_elem_id, sub_elem_val, fetch_parent)

        return sub_selected_element

    # pylint: disable=too-many-branches,too-many-locals,consider-using-f-string
    def select(
        self,
        root_element,
        search_path,
        fetch_parent=False) -> Any:
        """
        Get the json element through it's path. Returned object is either [dict, list or unicode].

        Search patterns:
          1. elem1.elem2
          2. elem1.elem2[index]
          3. elem1.elem2{property}
          4. elem1.elem2{property}[index]
          5. elem1.elem2{property<value>}
          6. elem1.elem2[index].elem3
          7. elem1.elem2{property}.elem3
          8. elem1.elem2{property<value>}.elem3
          9. elem1.elem2{property<value>}[index].elem3
         10. elem1.elem2{property<value>}.{property2<value2>}.elem3
        """

        self.pat_elem = '%s+' % self.jsonNameRe
        self.pat_sel_elem_val = '(%s)?((\\{(%s)(<(%s)>)?\\})+)(\\[(%s)\\])?' % (
            self.pat_elem, self.pat_elem, self.pat_elem, self.jsonArrayIndexRe)
        self.pat_sub_expr = '(\\{%s\\})' % self.pat_elem
        self.pat_sub_expr_val = '\\{(%s)(<(%s)>)?\\}' % (self.pat_elem, self.pat_elem)
        selected_element = root_element

        # pylint: disable=too-many-nested-blocks
        try:
            search_tokens = search_path.split(self.separator)
            for nextElement in search_tokens:
                if nextElement.find('{') >= 0:  # Next element has nested elements
                    parts = re.search(self.pat_sel_elem_val, nextElement)
                    sel_elem_id = parts.group(1)
                    sub_parts = parts.group(2)
                    elem_array_group = parts.group(7)
                    elem_array_index = parts.group(8)
                    sub_expressions = re.compile(self.pat_sub_expr).split(sub_parts)
                    if sel_elem_id and isinstance(selected_element, dict):
                        selected_element = selected_element.get(sel_elem_id)
                    if selected_element:
                        # Our first element is a list, so we will have to loop and find all the elements
                        # and sub expressions in it.
                        if isinstance(selected_element, list):
                            for nextInList in selected_element:
                                sub_selected_element = self.__find_in_subex__(
                                    sub_expressions, nextInList, self.pat_sub_expr_val, fetch_parent)
                                # It sub_selected_element is not null then we have found what we wanted.
                                if sub_selected_element:
                                    selected_element = sub_selected_element
                                    break
                        # Check if there are indexed elements.
                        if elem_array_group and elem_array_index and isinstance(selected_element, list):
                            selected_element = selected_element[int(elem_array_index)]

                elif nextElement.find('[') >= 0:  # Next element is indexed
                    pat_sel_elem_idx = f'({self.pat_elem})\\[({self.jsonArrayIndexRe})\\]'
                    parts = re.search(pat_sel_elem_idx, nextElement)
                    sub_elem_id = parts.group(1)
                    elem_array_index = parts.group(2)
                    # TODO Implement subarray like elem[0][1][2]
                    if sub_elem_id is not None and elem_array_index is not None:
                        el = selected_element.get(sub_elem_id)
                        if isinstance(el, list):
                            selected_element = el[int(elem_array_index)]
                else:  # Next element is simple
                    selected_element = selected_element.get(nextElement)
        except (AttributeError, IndexError):
            selected_element = None

        return selected_element



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   main.addons.appman.templates
      @file: tpl-main.py
   @created: Tue, 1 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import sys

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.tools.commons import dirname, get_path, read_version
from hspylib.modules.cli.application.application import Application
from hspylib.modules.cli.vt100.vt_utils import exit_app
from hspylib.modules.qt.qt_application import QtApplication
from hspylib.modules.qt.views.qt_view import QtView

HERE = get_path(__file__)


class Main(Application):
    """TODO"""

    class MainQtView(QtView):
        UI_FILE = 'main_qt_view.ui'

        def __init__(self):
            # Must come after the initialization above
            super().__init__(self.UI_FILE)
            self.configs = AppConfigs.INSTANCE
            self.setup_ui()

        def setup_ui(self) -> None:
            """Connect signals and startup components"""
            print(self.ui.lblHello, self.ui.bboxOkCancel)

    # The application version
    VERSION = read_version(f"{HERE}/.version")

    def __init__(self, app_name: str):
        # Invoke the super constructor without source_dir parameter to skip creation of log and properties
        super().__init__(app_name, self.VERSION, self.USAGE, source_dir=dirname(__file__))
        self.main_view = QtApplication(self.MainQtView)

    def _setup_parameters(self, *params, **kwargs) -> None:
        """Initialize application parameters and options"""

    def _main(self, *params, **kwargs) -> None:
        """Run the application with the command line arguments"""
        exit_app(self.main_view.run())

    def _cleanup(self) -> None:
        """Execute http_code cleanup before exiting"""


if __name__ == "__main__":
    # Application entry point
    Main('Application name').INSTANCE.run(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui
      @file: tui.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from abc import ABC, abstractmethod


class Menu(ABC):
    """TODO"""

    @abstractmethod
    def execute(self) -> 'Menu':
        """
        Execute the current tui actions.
        :return: The next tui action to proceed after processing.
        """

    @abstractmethod
    def trigger_menu_item(self) -> 'Menu':
        """
        Trigger the option action selected by the user.
        :return: The next tui action to trigger.
        """

    @abstractmethod
    def is_valid_option(self) -> bool:
        """
        Checks if the selected option is within the available menu_options.
        :return: whether the option is valid or not.
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   IP Message Sender. Sends TCP/UDP messages (multi-threaded).

   @project: HSPyLib
   @package: main.addons.widman.widgets
      @file: widget_send_msg.py
   @created: Thu, 26 Aug 2017
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
import os
import signal
import socket
import threading
from time import sleep
from typing import List

from hspylib.addons.widman.widget import Widget
from hspylib.core.enums.exit_code import ExitCode
from hspylib.core.exception.exceptions import WidgetExecutionError
from hspylib.core.tools.commons import sysout
from hspylib.modules.cli.application.argument_parser import HSArgumentParser
from hspylib.modules.cli.icons.font_awesome.widget_icons import WidgetIcons
from hspylib.modules.cli.tui.extra.minput.input_validator import InputValidator
from hspylib.modules.cli.tui.extra.minput.minput import MenuInput, minput
from hspylib.modules.cli.tui.menu.menu_utils import MenuUtils


class WidgetSendMsg(Widget):
    """HSPyLib to send TCP/UDP messages (multi-threaded)"""
    WIDGET_ICON = WidgetIcons.NETWORK
    WIDGET_NAME = "SendMsg"
    TOOLTIP = "IP Message Sender. Sends TCP/UDP messages (multi-threaded)"
    VERSION = (0, 3, 0)
    MAX_THREADS = 1000
    NET_TYPE_UDP = 'UDP'
    NET_TYPE_TCP = 'TCP'

    # Help message to be displayed by the application.
    USAGE = f"""Usage: SendMsg [options]

  Options:
    +n, ++net_type   <network_type>     : The network type to be used. Either udp or tcp ( default is tcp ).
    +p, ++port       <port_num>         : The port number [1-65535] ( default is 12345).
    +a, ++address    <host_address>     : The address of the datagram receiver ( default is 127.0.0.1 ).
    +k, ++packets    <num_packets>      : The number of max datagrams to be send. If zero is specified, then the app
                                          is going to send indefinitely ( default is 100 ).
    +i, ++interval   <interval_MS>      : The interval in seconds between each datagram ( default is 1 Second ).
    +t, ++threads    <threads_num>      : Number of threads [1-{MAX_THREADS}] to be opened to send simultaneously
                                          ( default is 1 ).
    +m, ++message    <message/filename> : The message to be sent. If the message matches a filename, then the file
                                          contents sent instead.

    E.g:. send-msg.py +n tcp +m "Hello" +p 12345 +a 0.0.0.0 +k 100 +i 500 +t 2
"""

    def __init__(self):
        super().__init__(
            self.WIDGET_ICON,
            self.WIDGET_NAME,
            self.TOOLTIP,
            self.USAGE,
            self.VERSION)

        self.is_alive = True
        self.net_type = None
        self.host = None
        self.packets = None
        self.interval = None
        self.threads = None
        self.message = None
        self.args = None
        self.socket = None
        self.counter = 1

    def execute(self, args: List[str] = None) -> ExitCode:
        signal.signal(signal.SIGINT, self.cleanup)
        signal.signal(signal.SIGTERM, self.cleanup)

        if args and args[0] in ['-h', '--help']:
            sysout(self.usage())
            return ExitCode.SUCCESS
        if args and args[0] in ['-v', '--version']:
            sysout(self.version())
            return ExitCode.SUCCESS

        if args and not self._parse_args(args):
            return ExitCode.ERROR
        elif not args and not self._read_args():
            return ExitCode.ERROR
        elif not args and not self.args:
            return ExitCode.ERROR

        self.net_type = self.args.net_type or self.NET_TYPE_TCP
        self.host = (self.args.address or '127.0.0.1', self.args.port or 12345)
        self.packets = self.args.packets or 100
        self.interval = self.args.interval or 1
        self.threads = self.args.threads or 1

        if self.args.message and os.path.isfile(self.args.message):
            file_size = os.stat(self.args.message).st_size
            sysout(f"Reading contents from file: {self.args.message} ({file_size}) [Bs] instead")
            with open(self.args.message, 'r', encoding='utf-8') as f_msg:
                self.message = f_msg.read()
        else:
            self.message = self.args.message or f"This is a {self.args.net_type} test %(count)"

        self._start_send()

        MenuUtils.wait_enter()

        return ExitCode.SUCCESS

    def cleanup(self) -> None:
        sysout('Terminating threads%NC%')
        self.is_alive = False
        if self.net_type == self.NET_TYPE_TCP:
            sysout('Closing TCP connection')
            self.socket.close()

    def _read_args(self) -> bool:
        """When no input is provided (e.g:. when executed from dashboard). Prompt the user for the info. """
        # @formatter:off
        form_fields = MenuInput.builder() \
            .field() \
                .label('Net Type') \
                .itype('select') \
                .value(f"{self.NET_TYPE_TCP}|{self.NET_TYPE_UDP}") \
                .build() \
            .field() \
                .label('Address') \
                .validator(InputValidator.anything()) \
                .value('127.0.0.1') \
                .build() \
            .field() \
                .label('Port') \
                .validator(InputValidator.numbers()) \
                .min_max_length(2, 5) \
                .value(12345) \
                .build() \
            .field() \
                .label('Packets') \
                .validator(InputValidator.numbers()) \
                .min_max_length(1, 4) \
                .value(100) \
                .build() \
            .field() \
                .label('Interval') \
                .validator(InputValidator.numbers()) \
                .min_max_length(1, 4) \
                .value(1) \
                .build() \
            .field() \
                .label('Threads') \
                .validator(InputValidator.numbers()) \
                .min_max_length(1, 4) \
                .value(1) \
                .build() \
            .field() \
                .label('Message') \
                .validator(InputValidator.anything()) \
                .min_max_length(1, 40) \
                .value('This is a test') \
                .build() \
            .build()
        # @formatter:on
        self.args = minput(form_fields)

        return len(self.args.__dict__) > 1 if self.args else False

    def _parse_args(self, args: List[str]):
        """When arguments are passed from the command line, parse them"""
        parser = HSArgumentParser(
            prog='sendmsg', prefix_chars="+", description='Sends TCP/UDP messages (multi-threaded)')
        parser.add_argument(
            '+n', '++net-type', action='store', type=str, choices=['udp', 'tcp'], default='tcp', required=False,
            help='The network type to be used. Either udp or tcp ( default is tcp )')
        parser.add_argument(
            '+a', '++address', action='store', type=str, default='127.0.0.1', required=False,
            help='The address of the datagram receiver ( default is 127.0.0.1 )')
        parser.add_argument(
            '+p', '++port', action='store', type=int, default=12345, required=False,
            help='The port number [1-65535] ( default is 12345)')
        parser.add_argument(
            '+k', '++packets', action='store', type=int, default=100, required=False,
            help='The number of max datagrams to be send. If zero is specified, then the app '
                 'is going to send indefinitely ( default is 100 ).')
        parser.add_argument(
            '+i', '++interval', action='store', type=float, default=1, required=False,
            help='The interval in seconds between each datagram ( default is 1 Second )')
        parser.add_argument(
            '+t', '++threads', action='store', type=int, default=1, required=False,
            help=f'Number of threads [1-{self.MAX_THREADS}] to be opened to send simultaneously ( default is 1 )')
        parser.add_argument(
            '+m', '++message', action='store', type=str, required=False,
            help='The message to be sent. If the message matches a filename, then the file contents sent instead')
        self.args = parser.parse_args(args)

        return bool(self.args)

    def _init_sockets(self) -> None:
        """Initialize sockets"""
        if self.net_type == self.NET_TYPE_UDP:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                self.socket.connect(self.host)
                sysout(f"Successfully connected to {self.host}")
            except socket.error as err:
                raise WidgetExecutionError('Unable to initialize sockets') from err

    def _start_send(self) -> None:
        """Start sending packets"""
        thread_relief = 0.05
        self._init_sockets()
        sysout(f"\n%ORANGE%Start sending {self.packets} "
               f"{self.net_type.upper()} packet(s) "
               f"every {self.interval} second(s) to {self.host} using {self.threads} thread(s)")
        threads_num = threading.active_count()

        for thread_num in range(1, int(self.threads) + 1):
            tr = threading.Thread(target=self._send_packet, args=(thread_num,))
            tr.setDaemon(True)
            tr.start()
            sleep(thread_relief)

        while self.is_alive and threading.active_count() > threads_num:
            sleep(2 * thread_relief)

    def _send_packet(self, thread_num: int) -> None:
        """Send a packet"""
        lock = threading.Lock()

        while self.is_alive and self.packets <= 0 or self.counter <= self.packets:
            message = self.message.replace('%(count)', str(self.counter))
            length = len(message)
            sysout(f"%BLUE%[Thread-{thread_num:d}] "
                   f"%GREEN%Sending \"{message:s}\" ({length:d}) bytes, "
                   f"Pkt = {self.counter:>d}/{self.packets:>d} %NC%...")
            if self.net_type == self.NET_TYPE_UDP:
                self.socket.sendto(message.encode(), self.host)
            else:
                try:
                    self.socket.sendall((message + '\n').encode())
                    with lock:
                        self.counter += 1
                except socket.error as err:
                    raise WidgetExecutionError('Unable to send packet') from err
            sleep(self.interval)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.vt100
      @file: vt_utils.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import os
import re
import signal
import sys
import termios
import tty
from typing import List, Optional, Tuple

from hspylib.core.tools.commons import is_debugging, sysout
from hspylib.modules.cli.vt100.vt_100 import Vt100
from hspylib.modules.cli.vt100.vt_codes import vt_print
from hspylib.modules.cli.vt100.vt_colors import VtColors


def require_terminal():
    if not sys.stdin.isatty():
        raise Exception('This module requires a terminal (TTY)')


def screen_size() -> Optional[List[str]]:
    """Retrieve the size of the terminal"""
    if sys.stdout.isatty():
        return os.popen('stty size').read().split()

    return None


# Solution taken from:
# https://stackoverflow.com/questions/46651602/determine-the-terminal-cursor-position-with-an-ansi-sequence-in-python-3
def get_cursor_position() -> Optional[Tuple[int, int]]:
    """ Get the terminal cursor position """
    if sys.stdout.isatty() and not is_debugging():
        buf = ""
        stdin = sys.stdin.fileno()
        attrs = termios.tcgetattr(stdin)
        try:
            tty.setcbreak(stdin, termios.TCSANOW)
            sys.stdout.write("\x1b[6n")
            sys.stdout.flush()
            while True:
                buf += sys.stdin.read(1)
                if buf[-1] == "R":
                    break
        finally:
            termios.tcsetattr(stdin, termios.TCSANOW, attrs)
        try:
            matches = re.match(r"^\x1b\[(\d*);(\d*)R", buf)
            groups = matches.groups()
        except AttributeError:
            return None
        return int(groups[0]), int(groups[1])

    return None


def set_enable_echo(enabled: bool = True) -> None:
    """Enable echo in the terminal
    :param enabled: whether is enabled or not
    """
    if sys.stdout.isatty():
        os.popen(f"stty {'echo -raw' if enabled else 'raw -echo min 0'}").read()


def set_auto_wrap(auto_wrap: bool = True) -> None:
    """Set auto-wrap mode in the terminal
    :param auto_wrap: whether auto_wrap is set or not
    """
    vt_print(Vt100.set_auto_wrap(auto_wrap))


def set_show_cursor(show_cursor: bool = True):
    """Show or hide cursor in the terminal
    :param show_cursor: whether to show or hide he cursor
    """
    vt_print(Vt100.set_show_cursor(show_cursor))


def save_cursor():
    """Save cursor position and attributes"""
    vt_print(Vt100.save_cursor())


def restore_cursor():
    """Restore cursor position and attributes"""
    vt_print(Vt100.restore_cursor())


def restore_terminal(clear_screen: bool = True):
    """Clear terminal and restore default attributes"""
    if clear_screen:
        vt_print('%HOM%%ED2%%MOD(0)%')
    set_auto_wrap()
    set_show_cursor()
    set_enable_echo()
    sysout('%NC%')


def exit_app(exit_code: int = signal.SIGHUP, frame=None, exit_msg: str = "Done.") -> None:
    """Exit the application. Commonly hooked to signals"""
    sysout(str(frame) if frame else '', end='')
    sysout(f"%HOM%%ED2%%NC%\n{exit_msg}\n")
    restore_terminal(False)
    sys.exit(exit_code if exit_code else 0)


def prepare_render(render_msg: str = '', render_color: VtColors = VtColors.ORANGE):
    """Prepare the terminal for TUI renderization"""
    signal.signal(signal.SIGINT, exit_app)
    signal.signal(signal.SIGHUP, exit_app)
    set_auto_wrap(False)
    set_show_cursor(False)
    sysout(f"%ED2%%HOM%{render_color.placeholder()}{render_msg}%HOM%%CUD(1)%%ED0%")
    save_cursor()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   @Package: main.enums
      @file: exit_code.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration


class ExitCode(Enumeration):
    """TODO"""

    # @formatter:off

    # Returned when something went wrong due to any Human interaction
    ERROR   = -1

    # Returned when something ran successfully without errors
    SUCCESS = 0

    # Returned when something that was supposed to work and failed due to unexpected software behaviour
    FAILED  = 1

    # @formatter:on

    def __str__(self):
        return "{}{}({})".format(
            '\033[0;32m' if self.name == 'SUCCESS' else '\033[0;31m',
            self.name, self.value
        )



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.31
#
# Package: main.hspylib.modules.cli.tui.menu
"""Package initialization."""

__all__ = [
  'menu', 
  'menu_entry', 
  'menu_factory', 
  'menu_item', 
  'menu_option', 
  'menu_ui', 
  'menu_utils'
]
__version__ = '0.11.31'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.db.firebase
      @file: firebase_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import base64
import logging as log
import os
import uuid

from requests.structures import CaseInsensitiveDict

from hspylib.core.config.app_config import AppConfigs
from hspylib.core.enums.charset import Charset
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.tools.preconditions import check_argument, check_state
from hspylib.modules.fetch.fetch import get


class FirebaseConfig(metaclass=Singleton):
    """TODO"""

    CONFIG_FORMAT = """
# Your Firebase configuration:
# --------------------------
PROJECT_ID={}
USERNAME={}
DATABASE={}
PASSPHRASE={}
UUID={}
"""

    @staticmethod
    def of(config_dict: CaseInsensitiveDict) -> 'FirebaseConfig':
        """TODO"""
        return FirebaseConfig(
            config_dict['PROJECT_ID'],
            config_dict['DATABASE'],
            config_dict['UUID'] if 'UUID' in config_dict else None,
            config_dict['USERNAME'],
            config_dict['PASSPHRASE'],
        )

    @staticmethod
    def of_file(filename: str, encoding: str = 'utf-8') -> 'FirebaseConfig':
        """TODO"""
        check_argument(os.path.exists(filename), "Config file does not exist")

        with open(filename, encoding=encoding) as f_config:
            cfg = CaseInsensitiveDict()
            for line in f_config:
                line = line.strip()
                if line.startswith("#") or "=" not in line:
                    continue
                key, value = line.split("=", 1)
                cfg[key.strip()] = value.strip()
            check_state(
                len(cfg) > 4,
                "Invalid configuration file. Must include at least: [PROJECT_ID, FIREBASE_URL, USERNAME, PASSPHRASE]")

            return FirebaseConfig.of(cfg) if len(cfg) == 5 else None

    def __init__(
        self,
        project_id: str = None,
        database: str = None,
        project_uuid: str = None,
        username: str = None,
        passphrase: str = None):

        self.encoding = str(Charset.UTF_8).lower()
        self.current_state = None
        self.project_id = project_id if project_id else AppConfigs.INSTANCE['firebase.project.id']
        self.database = database if database else AppConfigs.INSTANCE['firebase.database']
        self.project_uuid = project_uuid if project_uuid else AppConfigs.INSTANCE['firebase.project.uuid']
        self.username = username if username else AppConfigs.INSTANCE['firebase.username']
        self.passphrase = passphrase if passphrase else AppConfigs.INSTANCE['firebase.passphrase']
        check_state(self.project_id, "Project ID must be defined")
        check_state(self.database, "Database name must be defined")
        check_state(self.username, "Username must be defined")
        self.set_passphrase()
        self.project_uuid = str(self.project_uuid) if self.project_uuid else str(uuid.uuid4())
        check_state(self.validate_config(), "Your Firebase configuration is not valid")
        log.debug("Successfully connected to Firebase: %s", self.base_url())

    def __str__(self):
        return self.CONFIG_FORMAT.format(
            self.project_id,
            self.username,
            self.database,
            str(base64.b64encode(self.passphrase.encode(self.encoding)), encoding=self.encoding),
            self.project_uuid
        )

    def set_passphrase(self) -> None:
        """TODO"""
        check_state(
            self.passphrase and len(self.passphrase) >= 8,
            "Passphrase must be have least 8 characters size and must be base64 encoded")
        self.passphrase = str(base64.b64decode(self.passphrase), encoding=self.encoding)

    def validate_config(self) -> bool:
        """TODO"""
        response = get(f'{self.base_url()}.json')
        is_valid = response is not None and response.status_code == HttpCode.OK
        if is_valid:
            self.current_state = response.body if response.body and response.body != 'null' else None
        return is_valid

    def base_url(self) -> str:
        """TODO"""
        return f'https://{self.project_id}.firebaseio.com/{self.database}/{self.project_uuid}'

    def url(self, db_alias: str) -> str:
        """TODO"""
        return f'{self.base_url()}/{db_alias}.json'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.file
      @file: file_storage.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import ast
import logging as log
import os


class FileStorage:
    """TODO"""

    def __init__(self, filename: str):
        self.filename = filename
        self.data = []
        self.load()
        log.debug("File storage filename=%s created and loaded entries=%d", self.filename, len(self.data))

    def load(self) -> None:
        """TODO"""
        mode = 'r+' if os.path.exists(self.filename) else 'w+'
        with open(self.filename, mode, encoding='utf-8') as f_local_db:
            lines = f_local_db.read()
            if lines:
                saved_data = ast.literal_eval(lines)
                self.data = saved_data
        log.debug("File storage filename=%s loaded %d entries", self.filename, len(self.data))

    def commit(self) -> None:
        """TODO"""
        with open(self.filename, 'w', encoding='utf-8') as f_local_db:
            f_local_db.write(str(self.data))
        log.debug("File storage filename=%s committed entries=%d", self.filename, len(self.data))

    def truncate(self) -> None:
        """TODO"""
        with open(self.filename, 'w', encoding='utf-8'):
            log.warning("File storage filename=%s was truncated", self.filename)



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.31
#
# Package: main.hspylib.addons.appman.templates
"""Package initialization."""

__all__ = [
  'tpl-main', 
  'tpl-main-qt', 
  'tpl-test_main', 
  'tpl-widget'
]
__version__ = '0.11.31'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   main.addons.appman.templates
      @file: tpl-widget.py
   @created: Tue, 1 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from addons.widman.widget import Widget

from hspylib.core.enums.exit_code import ExitCode
from hspylib.core.tools.commons import sysout
from hspylib.modules.cli.icons.font_awesome.widget_icons import WidgetIcons
from hspylib.modules.cli.tui.menu.menu_utils import MenuUtils


class Widget_WIDGET_NAME_(Widget):
    WIDGET_ICON = WidgetIcons.WIDGET
    WIDGET_NAME = "_WIDGET_NAME_"
    TOOLTIP = "TODO usage."
    USAGE = "Usage: _WIDGET_NAME_"
    VERSION = (0, 1, 0)

    def __init__(self):
        super().__init__(
            self.WIDGET_ICON,
            self.WIDGET_NAME,
            self.TOOLTIP,
            self.USAGE,
            self.VERSION)

    def execute(self, *args) -> ExitCode:
        # Include the widget's main http_code
        sysout('')
        sysout('My widget is running')
        sysout('')

        MenuUtils.wait_enter()

        return ExitCode.SUCCESS

    def cleanup(self) -> None:
        # If your widget requires any cleanup procedures
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @Package: main.crud.file
      @file: file_repository.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import logging as log
import re
import uuid
from abc import abstractmethod
from typing import Any, List, Optional

from hspylib.core.crud.crud_entity import CrudEntity
from hspylib.core.crud.crud_repository import CrudRepository
from hspylib.core.crud.file.file_storage import FileStorage
from hspylib.core.exception.exceptions import ProgrammingError


class FileRepository(CrudRepository):
    """TODO"""

    _storages = {}

    @staticmethod
    def check_criteria(partial_value, whole_value) -> bool:
        """TODO
        :param partial_value:
        :param whole_value:
        :return:
        """
        if isinstance(whole_value, str):
            return str(partial_value).upper() in whole_value.upper()
        if isinstance(whole_value, int):
            return int(partial_value) == whole_value
        if isinstance(whole_value, float):
            return float(partial_value) == whole_value
        if isinstance(whole_value, bool):
            return bool(partial_value) == whole_value

        return False

    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.storage = self.__get_storage()

    def __str__(self):
        return str(self.storage.data)

    def insert(self, entity: CrudEntity) -> None:
        """TODO
        :param entity:
        """
        entity.uuid = entity.uuid if entity.uuid else str(uuid.uuid4())
        self.storage.data.append(entity.to_dict())
        self.storage.commit()
        log.debug("%s has been inserted !", entity.__class__.__name__)

    def update(self, entity: CrudEntity) -> None:
        """TODO
        :param entity:
        """
        for index, next_entry in enumerate(self.storage.data):
            if next_entry['uuid'] == entity.uuid:
                self.storage.data[index] = entity.to_dict()
                self.storage.commit()
                log.debug("%s has been updated !", entity.__class__.__name__)

    def delete(self, entity: CrudEntity) -> None:
        """TODO
        :param entity:
        """
        for index, next_entry in enumerate(self.storage.data):
            if next_entry['uuid'] == entity.uuid:
                self.storage.data.remove(self.storage.data[index])
                self.storage.commit()
                log.debug("%s has been deleted !", entity.__class__.__name__)

    def find_all(self, filters: str = None) -> List[CrudEntity]:
        """TODO
        :param filters:
        """
        self.storage.load()
        if filters:
            file_filters = filters.split(',')
            filtered = []
            for next_filter in file_filters:
                fields = re.split('=|>|<|>=|<=|==|!=', next_filter)
                try:
                    found = [
                        self.dict_to_entity(data) for data in self.storage.data if
                        self.check_criteria(fields[1], data[fields[0]])
                    ]
                except KeyError:
                    continue  # Just skip the filter and continue
                except IndexError:
                    continue  # Just skip the filter and continue
                filtered.extend(found)
            return filtered

        return [self.dict_to_entity(data) for data in self.storage.data]

    def find_by_id(self, entity_id: uuid.UUID) -> Optional[CrudEntity]:
        """TODO
        :param entity_id:
        """
        self.storage.load()
        if entity_id:
            result = [data for data in self.storage.data if entity_id == data['uuid']]
            if len(result) > 1:
                raise ProgrammingError(f'Multiple results {len(result)} found with entity_id={entity_id}')
            return self.dict_to_entity(result[0]) if len(result) > 0 else None

        return None

    @abstractmethod
    def dict_to_entity(self, row: dict) -> CrudEntity:
        """TODO
        :param row:
        """

    def __get_storage(self) -> Any:
        """Create or get a storage according to the filename"""
        if self.filename in self._storages:
            return self._storages[self.filename]

        self._storages[self.filename] = FileStorage(self.filename)
        return self._storages[self.filename]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui.extra.minput
      @file: field_builder.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
from typing import Any, TypeVar

from hspylib.core.tools.commons import str_to_bool
from hspylib.core.tools.constants import TRUE_VALUES
from hspylib.core.tools.preconditions import check_argument
from hspylib.core.tools.text_tools import camelcase
from hspylib.modules.cli.tui.extra.minput.access_type import AccessType
from hspylib.modules.cli.tui.extra.minput.form_field import FormField
from hspylib.modules.cli.tui.extra.minput.input_type import InputType
from hspylib.modules.cli.tui.extra.minput.input_validator import InputValidator
from hspylib.modules.cli.tui.extra.minput.minput_utils import MInputUtils

T = TypeVar('T')


class FieldBuilder:
    """TODO"""

    def __init__(self, parent: Any):
        self.parent = parent
        self.field = FormField()

    def label(self, label: str) -> 'FieldBuilder':
        self.field.label = label
        return self

    def itype(self, itype: str) -> 'FieldBuilder':
        self.field.itype = InputType.of_value(itype)
        return self

    def validator(self, validator: InputValidator) -> 'FieldBuilder':
        self.field.validator = validator
        return self

    def min_max_length(self, min_length: int, max_length: int) -> 'FieldBuilder':
        check_argument(max_length >= min_length, "Not a valid field length: ({}-{})", min_length, max_length)
        check_argument(max_length > 0 and min_length > 0, "Not a valid field length: ({}-{})", min_length, max_length)
        self.field.min_length = min_length
        self.field.max_length = max_length
        return self

    def access_type(self, access_type: str) -> 'FieldBuilder':
        self.field.access_type = AccessType.of_value(access_type)
        return self

    def value(self, value: T) -> 'FieldBuilder':
        check_argument(self.field.assign(value),
                       "Not a valid value: \"{}\". Validation pattern=\"{}\"", value, self.field.validator)
        return self

    def build(self) -> Any:
        self.field.itype = self.field.itype or InputType.TEXT
        if self.field.itype == InputType.CHECKBOX:
            self.field.value = '1' if str_to_bool(str(self.field.value)) else '0'
            self.field.min_length = self.field.max_length = 1
            self.validator(InputValidator.custom(r'[01]'))
        elif self.field.itype == InputType.SELECT:
            self.field.min_length = self.field.max_length = 1
            self.validator(InputValidator.anything())
        elif self.field.itype == InputType.MASKED:
            _, mask = MInputUtils.unpack_masked(self.field.value)
            self.field.min_length = self.field.max_length = len(mask)
            self.validator(InputValidator.custom(mask.replace('#', '[0-9]').replace('@', '[a-zA-Z]').replace('*', '.')))
        self.field.label = camelcase(self.field.label) or 'Field'
        self.field.min_length = self.field.min_length or 1
        self.field.max_length = self.field.max_length or 30
        self.field.access_type = self.field.access_type or AccessType.READ_WRITE
        self.field.icon = self.field.get_icon()
        self.field.value = self.field.value if self.field.value else ''
        self.parent.fields.append(self.field)
        return self.parent



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: main.modules.cli.tui.extra.minput
      @file: minput_utils.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
import re
from abc import ABC
from typing import Any, Optional, Tuple

from hspylib.core.exception.exceptions import InvalidInputError
from hspylib.core.tools.preconditions import check_argument
from hspylib.modules.cli.vt100.vt_codes import vt_print


class MInputUtils(ABC):
    """TODO"""

    @staticmethod
    def detail_len(field: Any) -> int:
        max_len = len(str(field.max_length))
        return 1 + (2 * max_len)

    @staticmethod
    def mi_print(size: int = 0, text: str = None, prepend: str = None, end: str = '') -> None:
        fmt = ('{}' if prepend else '') + "{:<" + str(size) + "} : "
        if prepend:
            vt_print(fmt.format(prepend, text), end=end)
        else:
            vt_print(fmt.format(text or ''), end=end)

    @staticmethod
    def toggle_selected(tokenized_values: str) -> str:
        values = tokenized_values.split('|')
        cur_idx = next((idx for idx, val in enumerate(values) if val.find('<') >= 0), -1)
        if cur_idx < 0:
            if len(values) > 1:
                values[1] = f'<{values[1]}>'
            else:
                values[0] = f'<{values[0]}>'
            return '|'.join(values)

        unselected = list(map(lambda x: x.replace('<', '').replace('>', ''), values))
        # @formatter:off
        return '|'.join([
            f'<{val}>'
            if
                idx == (cur_idx + 1)
                or ((cur_idx + 1) >= len(unselected) and idx == 0)
            else
                val for idx, val in enumerate(unselected)
        ])
        # @formatter:on

    @staticmethod
    def get_selected(tokenized_values: str) -> Optional[Tuple[int, str]]:
        values = tokenized_values.split('|')
        # @formatter:off
        sel_item = next((
            val.replace('<', '').replace('>', '')
            for val in values if val.startswith('<') and val.endswith('>')
        ), values[0])
        # @formatter:on
        try:
            return values.index(sel_item), sel_item
        except ValueError:
            try:
                return values.index(f"<{sel_item}>"), sel_item
            except ValueError:
                return -1, sel_item

    @staticmethod
    def unpack_masked(value: str) -> Tuple[str, str]:
        parts = value.split('|')
        check_argument(len(parts) == 2, "Invalid masked value: {}", value)
        return parts[0], parts[1]

    @staticmethod
    def append_masked(value: str, mask: str, keypress_value: chr) -> str:
        masked_value = value
        idx = len(value)
        while idx < len(mask) and mask[idx] not in ['#', '@', '*']:
            masked_value += mask[idx]
            idx += 1
        if mask and idx < len(mask):
            mask_regex = mask[idx].replace('#', '[0-9]').replace('@', '[a-zA-Z]').replace('*', '.')
            if re.search(mask_regex, keypress_value):
                masked_value += keypress_value
            else:
                raise InvalidInputError(f"Value {keypress_value} is not a valid value against mask: {mask}")

        return f"{masked_value}|{mask}"

    @staticmethod
    def over_masked(value: str, mask: str) -> str:
        masked_value = ''
        for idx, element in enumerate(mask):
            if element in ['#', '@', '*']:
                masked_value += value[idx] if idx < len(value) else '_'
            else:
                masked_value += element

        return masked_value
